package org.embulk.input.filename;

import java.io.*;
import java.util.*;

import java.util.Arrays;
import java.util.Comparator;
import java.util.Date;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.nio.file.FileSystems;
import java.nio.file.attribute.FileTime;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.Files;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.FileVisitResult;
import java.nio.file.attribute.BasicFileAttributes;
import com.google.common.collect.ImmutableList;
import com.google.common.base.Optional;
import org.slf4j.Logger;
import org.embulk.config.Config;
import org.embulk.config.ConfigDefault;
import org.embulk.config.ConfigException;
import org.embulk.config.ConfigInject;
import org.embulk.config.ConfigSource;
import org.embulk.config.ConfigDiff;
import org.embulk.config.TaskReport;
import org.embulk.config.Task;
import org.embulk.config.TaskSource;
import org.embulk.spi.Exec;
import org.embulk.spi.FileInputPlugin;
import org.embulk.spi.BufferAllocator;
import org.embulk.spi.TransactionalFileInput;
import org.embulk.spi.util.InputStreamTransactionalFileInput;
import org.embulk.standards.LocalFileInputPlugin;



class FilenameFileInputStream extends FileInputStream {
    static int MAX_NAME_LENGTH = 255;
    int n;
    byte[] name;

    FilenameFileInputStream(File file) throws FileNotFoundException {
        super(file);
        n = 0;
        name = file.getName().getBytes();
    }

    FilenameFileInputStream(String path) throws FileNotFoundException {
        super(path);
        n = 0;
        name = path.getBytes();
    }

    @Override
    public int read() throws IOException {
        if (n < name.length) {
            byte b = name[n];
            n++;
            return b;
        } else if (n < MAX_NAME_LENGTH) {
            n++;
            return 0;
        } else {
            return super.read();
        }
    }

    @Override
    public int read(byte[] b) throws IOException {
        return read(b, 0, b.length);
    }

    @Override
    public int read(byte[] b, int off, int len) throws IOException {
        if (n < MAX_NAME_LENGTH) {
            int i = 0;
            int c;
            for (; i < len; i++) {
                c = read();
                if (c == -1) {
                    if ( i == 0 ) {
                        return -1;
                    }
                    break;
                }
                b[off + i] = (byte)c;
            }
            return i;
         } else {
            return super.read(b, off, len);
         }
    }
}

public class FilenameFileInputPlugin implements FileInputPlugin
{

    public interface PluginTask extends Task
    {
        @Config("path_prefix")
        @ConfigDefault("null")
        Optional <String> getPathPrefix();

        @Config("path_list")
        @ConfigDefault("null")
        Optional <List<String>> getPathList();

        @Config("file_name_suffix")
        @ConfigDefault("null")
        Optional <String> getFileNameSuffix();

        @Config("start_date")
        @ConfigDefault("null")
        Optional<String> getStartDate();
 
        @Config("first_path")
        @ConfigDefault("null")
        Optional<String> getFirstPath();

        @Config("last_path")
        @ConfigDefault("null")
        Optional<String> getLastPath();

        @Config("file_size")
        @ConfigDefault("null")
        Optional<Integer> getFileSize();

        @Config("max_file_count")
        @ConfigDefault("null")
        Optional<Integer> getMaxFileCount();

        @Config("follow_symlinks")
        @ConfigDefault("false")
        boolean getFollowSymlinks();

        @Config("use_last_modified")
        @ConfigDefault("true")
        boolean getUseLastModified();


        @Config("last_modified")
        @ConfigDefault("null")
        Optional<Long> getLastModified();


        List<String> getFiles();
        void setFiles(List<String> files);

        @ConfigInject
        BufferAllocator getBufferAllocator();
    }

    private final Logger log = Exec.getLogger(getClass());

    private final static Path CURRENT_DIR = Paths.get(".").normalize();

    @Override
    public ConfigDiff transaction(ConfigSource config, FileInputPlugin.Control control)
    {
        PluginTask task = config.loadConfig(PluginTask.class);

        // validate path_prefix: and path_list:
        if (task.getPathPrefix().isPresent() && task.getPathList().isPresent()) {
            throw new ConfigException("path_prefix: and path_list: must not be multi-select");
        }
        if (!task.getPathPrefix().isPresent() && !task.getPathList().isPresent()) {
            throw new ConfigException("Must require path_prefix: or path_list:");
        }

        // list files recursively
        List<String> files = new ArrayList<String>(listFiles(task));

	      // only process <= max_file_count number of files
	     final Integer maxFileCount = task.getMaxFileCount().orNull();

        if(task.getUseLastModified()) {
          File [] testFiles = new File[files.size()] ;
	        List<File> dateFileList = new ArrayList<>();
	        for(int i = 0; i< files.size(); i++){
            dateFileList.add(new File(files.get(i)));
			      testFiles[i] = new File(files.get(i));
          }

          List<String>  dateSortedFiles = new ArrayList<String>();
          for(File f: dateFileList) {
            dateSortedFiles.add(f.toString());
          }

	  Collections.sort(dateSortedFiles);
          task.setFiles(dateSortedFiles);

          // number of processors is same with number of files
          log.info("Loading files {}", dateSortedFiles);
          int taskCount = task.getFiles().size();
          return resume(task.dump(), taskCount, control);
        }

        else {
          Collections.sort(files);
	        if (maxFileCount != null && files.size() > maxFileCount) {
            files = files.subList(0, maxFileCount);
          }
	        log.info("Loading files {}", files);
	        task.setFiles(files);

          // number of processors is same with number of files
          int taskCount = task.getFiles().size();
          return resume(task.dump(), taskCount, control);
        }

    }

    @Override
    public ConfigDiff resume(TaskSource taskSource,
            int taskCount,
            FileInputPlugin.Control control)
    {
        PluginTask task = taskSource.loadTask(PluginTask.class);

        control.run(taskSource, taskCount);

        // build next config
        ConfigDiff configDiff = Exec.newConfigDiff();

        // last_path
        if (task.getFiles().isEmpty()) {
            // keep the last value
            if (task.getLastPath().isPresent()) {
                configDiff.set("last_path", task.getLastPath().get());
            }
	          if(task.getUseLastModified()){
	              if (task.getLastModified().isPresent()) {
                    configDiff.set("last_modified", task.getLastModified().get());
                }
	          }
        } else {
            List<String> files = new ArrayList<String>(task.getFiles());
            Collections.sort(files);
            configDiff.set("last_path", files.get(files.size() - 1));

            if(task.getUseLastModified()) {

              long last = Long.MIN_VALUE;
              long lastFileTime = Long.MIN_VALUE;
	            int  index = 0;
	            List sameTime = new ArrayList();

              for (int i = 0; i < files.size(); i++) {
             	    File f = new File(files.get(i));
	                Path f_name = Paths.get(f.getPath());

                  try {
                      FileTime ctime = (FileTime) Files.getAttribute(f_name, "unix:ctime");
		                  long ctime_long = ctime.toMillis();
		                  if  (ctime_long >= last){
                        last = ctime_long;
                        index = i;
                      }

                  } catch (IOException e) {
                      log.info("Cannot get the unix:ctime - {} ", e);
                  }
              }
		          File f = new File(files.get(index));
	            Path f_name = Paths.get(f.getPath());
              try {
                  FileTime ctime = (FileTime) Files.getAttribute(f_name, "unix:ctime");
		              configDiff.set("last_modified", ctime.toMillis());
              } catch (IOException e) {
                  log.info("Cannot get the unix:ctime - {} ", e);
              }
           }
      }

        return configDiff;
}
    @Override
    public void cleanup(TaskSource taskSource,
            int taskCount,
            List<TaskReport> successTaskReports)
    { }

    public List<String> listFiles(PluginTask task)
    {
        final ImmutableList.Builder<String> builder = ImmutableList.builder();

        if (task.getPathList().isPresent()) {
          Path p;
          for (String path_name : task.getPathList().get()) {
              p = Paths.get(path_name).normalize();
              if (Files.exists(p)) {
                  builder.add(p.toString());
              }
          }
        return builder.build();
        } 

        Path pathPrefix = Paths.get(task.getPathPrefix().orNull()).normalize();
        final Path directory;
        final String fileNamePrefix;

        if (Files.isDirectory(pathPrefix)) {
            directory = pathPrefix;
            fileNamePrefix = "";
        } else {
            fileNamePrefix = pathPrefix.getFileName().toString();
            Path d = pathPrefix.getParent();
            directory = (d == null ? CURRENT_DIR : d);
        }
        
        final SimpleDateFormat format = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss zzz");
        final String startDateString = task.getStartDate().orNull();
        final Date startDate;
        if (startDateString != null) {
            try {
                startDate = format.parse(startDateString);
            } catch (ParseException e) {
                throw new RuntimeException(String.format("Cannot parse start_date '%s'", startDateString), e);
            }
        } else {
            startDate = null;
        }
        final String firstPath = task.getFirstPath().orNull();
        final String lastPath = task.getLastPath().orNull();
        final Integer fileSize = task.getFileSize().orNull();
        final String fileNameSuffix = task.getFileNameSuffix().orNull();

        final Long lastModified;
        final boolean useLastModified = task.getUseLastModified();
        if (useLastModified) { 
            if (task.getLastModified().orNull() != null) {
                lastModified = task.getLastModified().orNull();
            } else if (lastPath != null) {
                FileTime lastPathTime;
                try {
                    lastPathTime = (FileTime) Files.getAttribute(Paths.get(lastPath), "unix:ctime");
                    lastModified = lastPathTime.toMillis();
                } catch (IOException e) {
                    throw new RuntimeException(String.format("Cannot get the last changed time of lastpath '%s'", lastPath), e);
                }
            } else {
              lastModified = null;
            }
        } else {
            lastModified = null;
        }
        try {
            log.info("Listing local files at directory '{}' filtering filename by prefix '{}'", directory.equals(CURRENT_DIR) ? "." : directory.toString(), fileNamePrefix);

            Files.walkFileTree(directory, new SimpleFileVisitor<Path>() {
                @Override
                public FileVisitResult preVisitDirectory(Path path, BasicFileAttributes attrs)
                {
	                if(useLastModified) {
		                FileTime fileTime;
	                        long fileChangeTime = Long.MIN_VALUE;
                        	try {
  	                                fileTime = (FileTime) Files.getAttribute(path, "unix:ctime");
                           		fileChangeTime = fileTime.toMillis();

        	                } catch (IOException e) {
                	           	log.info("Cannot get the last changed time - {} ", e);
                        	}
//				log.info("########################FROM REVISIT#############################################");
//				log.info ("file = {}", path);
//				log.info("last_Modified = {},    files time = {}", lastModified, fileChangeTime);
	                        if (path.equals(directory) || Files.isDirectory(path)) {
        	                    return FileVisitResult.CONTINUE;
                                } else if (firstPath != null && path.toString().compareTo(firstPath.substring(0, path.toString().length())) < 0) {
                                    return FileVisitResult.SKIP_SUBTREE;
                	        } else if (lastModified != null && lastModified > fileChangeTime) { //if file is older then lastModified
	                            return FileVisitResult.SKIP_SUBTREE;
        	                } else if (path.getFileName().toString().startsWith(".")) {
                	            return FileVisitResult.SKIP_SUBTREE;
                        	} else {
                            		if (path.getFileName().toString().startsWith(fileNamePrefix)) {
		                                return FileVisitResult.CONTINUE;
                	            	} else {
		                                return FileVisitResult.SKIP_SUBTREE;
                			}
                        	}
                   	}

	                 else {
    		               if (path.equals(directory)) {
                           return FileVisitResult.CONTINUE;
                       } else if (firstPath != null && path.toString().compareTo(firstPath.substring(0, path.toString().length())) < 0) {
                           return FileVisitResult.SKIP_SUBTREE;
                       } else if (lastPath != null && path.toString().compareTo(lastPath.substring(0, path.toString().length())) < 0) {
                           return FileVisitResult.SKIP_SUBTREE;
                       } else if (path.getFileName().toString().startsWith(".")) {
                           return FileVisitResult.SKIP_SUBTREE;
                       } else {
                           if (path.getFileName().toString().startsWith(fileNamePrefix)) {
                               return FileVisitResult.CONTINUE;
                           } else {
                               return FileVisitResult.SKIP_SUBTREE;
                           }
                       }
		              } //end_else lastPath
              }

              @Override
              public FileVisitResult visitFile(Path path, BasicFileAttributes attrs)
              {
                    FileTime fileModifiedTime;
                    Date fileModifiedDate = null;
                    try {
                        fileModifiedTime = (FileTime) Files.getLastModifiedTime(path);
                        fileModifiedDate = new Date(fileModifiedTime.toMillis());
                    } catch (IOException e) {
                        log.info("Cannot get the last modified time - {} ", e);
                    }
                    if (firstPath != null && path.toString().compareTo(firstPath) <= 0) {
                        return FileVisitResult.CONTINUE;
                    }
                    if (startDate != null && startDate.compareTo(fileModifiedDate) > 0) {
                        return FileVisitResult.CONTINUE;
                    }
                    if(useLastModified) {
                       FileTime fileTime;
                       long fileChangeTime = Long.MIN_VALUE;
                       try {
                           fileTime = (FileTime) Files.getAttribute(path, "unix:ctime");
                           fileChangeTime = fileTime.toMillis();

                        } catch (IOException e) {
                            log.info("Cannot get the last changed time - {} ", e);
                        }

//			log.info("########################FROM VISIT#############################################");
//			log.info ("file = {}", path);
//                      log.info("last_Modified = {},    files time = {}", lastModified, fileChangeTime);

                        if (lastModified != null && lastModified > fileChangeTime) {
                            return FileVisitResult.CONTINUE;
                        } else if (lastModified != null && lastModified == fileChangeTime && lastPath != null && path.toString().compareTo(lastPath) <= 0 ) {
			    return FileVisitResult.CONTINUE;
			}
			  else if (path.getFileName().toString().startsWith(".")) {
                            return FileVisitResult.CONTINUE;
                        } else {
                            if (path.getFileName().toString().startsWith(fileNamePrefix)) {
                              if (fileNameSuffix == null || path.getFileName().toString().endsWith(fileNameSuffix)) {
                                if (fileSize == null || path.toFile().length() == fileSize) {
                                  builder.add(path.toString());
                                }
                              }
                          }
                          return FileVisitResult.CONTINUE;
                      }

		              }

		              else{
		                  if (lastPath != null && path.toString().compareTo(lastPath) <= 0) {
                          return FileVisitResult.CONTINUE;
                      } else if (path.getFileName().toString().startsWith(".")) {
                          return FileVisitResult.CONTINUE;
                      } else {
                          if (path.getFileName().toString().startsWith(fileNamePrefix)) {
                            if (fileNameSuffix == null || path.getFileName().toString().endsWith(fileNameSuffix)) {
                              if (fileSize == null || path.toFile().length() == fileSize) {
                                builder.add(path.toString());
                              }
                            }
                        }
                        return FileVisitResult.CONTINUE;
                    }
		            }
              }
            });
        } catch (IOException ex) {
            throw new RuntimeException(String.format("Failed get a list of local files at '%s'", directory), ex);
        }
        return builder.build();
    }

    @Override
    public TransactionalFileInput open(TaskSource taskSource, int taskIndex)
    {
        final PluginTask task = taskSource.loadTask(PluginTask.class);
        final String path = task.getFiles().get(taskIndex);

        return new InputStreamTransactionalFileInput(
                task.getBufferAllocator(),
                new InputStreamTransactionalFileInput.Opener() {
                    public InputStream open() throws IOException
                    {
                        return new FilenameFileInputStream(path);
                    }
                })
        {
            @Override
            public void abort()
            { }

            @Override
            public TaskReport commit()
            {
                return Exec.newTaskReport();
            }
        };
    }
}
